#!/usr/bin/perl -w
use strict;
use warnings;
use Getopt::Long;
use Bio::SearchIO;
use Bio::SeqIO;

################################################################################
#Name:				dbsubtract.pl
#Author:			Daniel Gaston
#Contact:			daniel.gaston@dal.ca
#Last Modified:		December 11th, 2007
#Purpose:			To sort out relevant blast hits (with a length greater than
#					or equal to a given length, and with a percent ID greater
#					than or equal to a given percentage) from a large,
#					concatenated blast file. These hits are then removed from
#					the target database to create a subtractive database.
#Version:			0.1-beta
#License:			This script is published under the GNU license and is freely
#					available for use and modification as long as original credit
#					is given to the author
#Improvements:		If you have suggestions or improvements for this program
#					please contact the author
#Disclaimer:		The author assumes no responsibility for the results generated
#					using this program
#
#Changelog:         December 11th, 2007 (Previous Version: November 6th, 2006)
#                   No longer need to perform annotation when doing a subtraction
#
#                   Blastall run is done when requested instead of needing to be
#                   performed seperately
#
#                   Can perform multiple subtractions if proper flag is given to
#                   run in multi-mode.
################################################################################

################################################################################
#								Constants
################################################################################

my $IDENT_SCORE = 99;			#Default Value
my $LENGTH = 0.9;				#Default Value: Ratio of Query and Hit

################################################################################
#								Variables
################################################################################

my $verbose = 0;
my $help = 0;
my $debug = 0;
my $multiple = 0;
my $annotate = "";
my $subtract = "";
my $input = "";
my $help_msg = "";
my @hits;
my $i = 0;
my $hit_out = "mito_hits.dat";
my $query_length;
my %annotate;

################################################################################
#							Command Line Parsing
################################################################################

GetOptions ('verbose'    =>   \$verbose,
			'help'       =>   \$help,
			'debug'      =>   \$debug,
			'ident=i'    =>   \$IDENT_SCORE,
			'length=i'   =>   \$LENGTH,
			'output=s'   =>   \$hit_out,
			'annotate=s' =>   \$annotate,
			'subtract=s' =>   \$subtract,
            'multi'      =>   \$multiple);

$help_msg =

<<HELP;

	This program is designed to aid in the construction of a subtractive
	database using blastall. The blastall program is run seperately (future
	fully functional version will run blastall from within this program) and the
	program works on the generated blastresult file.
		
	Use:
		
	perl dbsubtract [options] [blastresults file]
		
	Example:
		
	perl dbsubtract.pl -a query.fasta -s database.fasta blastresults.dat
		
	Options:
		
	-v	--verbose:	verbose output
		
	-h	--help:		prints this help message and exits
		
	-d	--debug:	enables debug statements, used for
				testing
						
	-i	--ident:	specifies the lower percent identity
				threshold (default: 99%)
						
	-a	--annotate:	uses annotation info from target to
				annotate an unannotated query database.
				User must supply the name of the fasta
				file to be annotated if they specify
				this option.
						
	-s	--subtract:	Create a subtractive database. If this
				flag is specified then the name of the
				fasta file to annotate must be given
						
	-l	--length:	specifies the minimum alignment length
				ratio (default: 1)
						
	-o	--output:	specifies file name for intermediate
				file of hits matching the specified
				criteria. (default: mito_hits.dat)

HELP

if(scalar(@ARGV) != 1 || $help){
	print $help_msg;
	exit();
} else {
	$input = $ARGV[0];
}

################################################################################
#								Main Program
################################################################################

my $report = new Bio::SearchIO(-format => 'blast', 
							   -file   => $input);

open(OUT, ">$hit_out") or die "Could not create $hit_out!\n";

RESULT: while( my $result = $report->next_result ) {
	$query_length = $result->query_length;
	my $hit = $result->next_hit;
    if(!$hit){
        print "No hits for ", $result -> query_name, "\n";
        print OUT "**********************************\n";
        print OUT "No hits for ", $result -> query_name, "\n";
        print OUT "**********************************\n";
        next RESULT;   
    }
	my $hsp = $hit -> next_hsp;
    my $query_name = $result -> query_name();
    
	if( ($hsp->length('total'))/$query_length >= $LENGTH ) {
		if ( $hsp->percent_identity >= $IDENT_SCORE ) {
					
			if($verbose || $debug){
				print "Query: $query_name",
                      "\t\tHit: ",       $hit->name, 
					  "\t\tLength: ",     $hsp->length('total'), " ", $hsp->length('total')/$query_length,
					  "\t\tPercent_id: ", $hsp->percent_identity, "\n";
			}
					
			print OUT "Query: $query_name",
                      "\t\tHit: ", $hit->name,
					  "\t\tLength: ",     $hsp->length('total'), " ", $hsp->length('total')/$query_length,
					  "\t\tPercent ID: ", $hsp->percent_identity, "\n";

			$i++;
					
			if($annotate){
				#annotate the unannotated query database
				$annotate{$hit->name} = $hit->description;
			}
			if($subtract){
				#do the database subtraction
				push @hits, $hit->name;
			}
					
		}else{
            print OUT "**********************************\n";
            print OUT "No appropriate hit found for $query_name (Percent Identiy)\n";
            print OUT "Hit: ", $hit->name,
					  "\t\tLength: ",     $hsp->length('total'), " ", $hsp->length('total')/$query_length,
					  "\t\tPercent ID: ", $hsp->percent_identity, "\n";
            print OUT "**********************************\n";
        }
    }else{
        print OUT "**********************************\n";
        print OUT "No appropriate hit found for $query_name (Length)\n";
        print OUT "Hit: ", $hit->name,
				  "\t\tLength: ",     $hsp->length('total'), " ", $hsp->length('total')/$query_length,
				  "\t\tPercent ID: ", $hsp->percent_identity, "\n";
        print OUT "**********************************\n";
    }
}
close(OUT);
open(DUP, ">duplicates.dat") or die "Could not create duplicates.dat!\n";
for(my $i = 0; $i < scalar(@hits); $i++){
    for(my $y = $i + 1; $y < scalar(@hits); $y++){
        if($hits[$i] eq $hits[$y]){
            print DUP "$hits[$i]\n";
            if($verbose){
                print "Duplicate: $hits[$i]\n";   
            }
        }
    }
}
close(DUP);

#Open the original target database fasta file and remove all of the definite
#positive hits in order to create the subtractive database
if($subtract){
	if($verbose || $debug){
		print "Now creating subtractive database\n";
	}
	#call subtraction subroutine
	subtract(\@hits, $subtract, $verbose);
}

#Open the query sequence for annotation
if($annotate){
	if($verbose || $debug){
		print "Now annotating unannotated query fasta file\n";
	}
	#call the annotation subroutine
	annotate(\%annotate, $annotate);
}

print "Finished!\n";

################################################################################
#									End Main
################################################################################


################################################################################
#								Subroutines
################################################################################

sub subtract {
    my ($subtract_ref, $file, $verb) = @_;
	my @subtract = @$subtract_ref;
	my $out_file = $file;
	my $flag = 1;
	
	$out_file =~ s/.fasta$//g;
	$out_file = "$out_file" . "_subtract.fasta";
	
	open(IN, "$file") or die "Could not open $file for reading!\n";
	open(OUT, ">$out_file") or die "Could not create $out_file!\n";
    my @removed;
	my $seq = Bio::SeqIO->new(-file => "$file",
							  -format => "fasta");
	while (my $seq = $seq->next_seq) {
		#compare the sequence name from here to the one in our array, if there is
		#a match than it needs to be subtracted and so not written to the new
		#file. If there is no match write it to the new database file
		my $seq_id = $seq->id();
        if($verb){
            print "Testing $seq_id for removal...\n";   
        }
        foreach(@subtract){
			if($seq_id eq $_){
                if($verb){
                    print "Match Found for $seq_id. Removed..\n";   
                }
				$flag = 0;
                push @removed, $_;
				last;
			}
		}
		if($flag){
			#write to new file
            if($verb){
                print "$seq_id not removed, printing to subtractive database...\n";  
            }
			print OUT ">$seq_id | " . $seq->desc() . "\n";
			print OUT $seq->seq() . "\n";
		}
		$flag = 1;
	}
    if(scalar(@subtract) != scalar(@removed)){
        my @sorted_subtract = sort(@subtract);
        my @sorted_removed = sort(@removed);
        print "Warning: Following sequence was not matched...\n";
        for(my $i = 0; $i < scalar(@subtract); $i++){
            if($sorted_subtract[$i] ne $sorted_removed[$i]){
                print "$sorted_subtract[$i]\n";
                last;
            }
        }
    }
	close(IN);
	close(OUT);
}

sub annotate {
	my $annotate_ref = $_[0];
	my %annotate = %$annotate_ref;
	my $file = $_[1];
	my $out_file = $file;
	
	$out_file =~ s/.fasta$//g;
	$out_file = "$out_file" . "_annotated.fasta";
	
	open(IN, "$file") or die "Could not open $file for reading!\n";
	open(OUT, ">$out_file") or die "Could not create $out_file!\n";
	my $seq = Bio::SeqIO->new(-file => "$file",
							  -format => "fasta");
	while (my $seq = $seq->next_seq) {
		#Cycle through all of the sequences in the unannotated fasta file
		#and find the correct entry in the hash for that sequence, use the annotation
		#to set the description of the sequence object and write the new object to the
		#file
		
		my $seq_id = $seq->id();
		print OUT ">$seq_id | $annotate{$seq_id}\n";
		print OUT $seq->seq() . "\n";
		
	}
	close(OUT);
	close(IN);
}

################################################################################
#								End Subroutines
################################################################################